@@ -40,6 +40,7 @@ class Q_QUICK3D_EXPORT QQuick3DBounds3
     Q_GADGET
     Q_PROPERTY(QVector3D minimum READ minimum CONSTANT)
     Q_PROPERTY(QVector3D maximum READ maximum CONSTANT)
+    QML_VALUE_TYPE(bounds)
 
 public:
     QVector3D minimum() const
@@ -222,7 +223,6 @@ private:
     QQuick3DInstancing *m_instancing = nullptr;
     QQuick3DNode *m_instanceRoot = nullptr;
     QMetaObject::Connection m_geometryConnection;
-    QMetaObject::Connection m_skeletonConnection;
     QMetaObject::Connection m_instancingConnection;
     float m_depthBias = 0.0f;
     bool m_castsShadows = true;
