@@ -5,7 +5,9 @@
 #include <QtQuick3DRuntimeRender/private/qssgrhieffectsystem_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrenderer_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrhiquadrenderer_p.h>
-#include <QtQuick3DRuntimeRender/private/qssgrendercontextcore_p.h>
+#include "qssgrendercontextcore.h"
+#include "qssgrendershadercodegenerator_p.h"
+#include <qtquick3d_tracepoints_p.h>
 
 #include <QtQuick3DUtils/private/qssgassert_p.h>
 
@@ -66,7 +68,8 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::getTexture(const QByteArray &bufferNa
                                                       const QSize &size,
                                                       QRhiTexture::Format format,
                                                       bool isFinalOutput,
-                                                      const QSSGRenderEffect *inEffect)
+                                                      const QSSGRenderEffect *inEffect,
+                                                      quint8 viewCount)
 {
     QSSGRhiEffectTexture *result = findTexture(bufferName);
     const bool gotMatch = result != nullptr;
@@ -92,7 +95,8 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::getTexture(const QByteArray &bufferNa
         m_textures.append(result);
     }
 
-    QRhi *rhi = m_sgContext->rhi();
+    const auto &rhiCtx = m_sgContext->rhiContext();
+    QRhi *rhi = rhiCtx->rhi();
     const bool formatChanged = result->texture && result->texture->format() != format;
     const bool needsRebuild = result->texture && (result->texture->pixelSize() != size || formatChanged);
 
@@ -101,7 +105,10 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::getTexture(const QByteArray &bufferNa
         flags |= QRhiTexture::UsedAsTransferSource;
 
     if (!result->texture) {
-        result->texture = rhi->newTexture(format, size, 1, flags);
+        if (viewCount >= 2)
+            result->texture = rhi->newTextureArray(format, viewCount, size, 1, flags);
+        else
+            result->texture = rhi->newTexture(format, size, 1, flags);
         result->texture->create();
     } else if (needsRebuild) {
         result->texture->setFlags(flags);
@@ -111,7 +118,10 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::getTexture(const QByteArray &bufferNa
     }
 
     if (!result->renderTarget) {
-        result->renderTarget = rhi->newTextureRenderTarget({ result->texture });
+        QRhiColorAttachment colorAttachment(result->texture);
+        colorAttachment.setMultiViewCount(viewCount);
+        QRhiTextureRenderTargetDescription desc(colorAttachment);
+        result->renderTarget = rhi->newTextureRenderTarget(desc);
         result->renderPassDescriptor = result->renderTarget->newCompatibleRenderPassDescriptor();
         result->renderTarget->setRenderPassDescriptor(result->renderPassDescriptor);
         result->renderTarget->create();
@@ -151,12 +161,13 @@ void QSSGRhiEffectSystem::releaseTextures()
         releaseTexture(t);
 }
 
-QRhiTexture *QSSGRhiEffectSystem::process(const QSSGRenderEffect &firstEffect,
+QRhiTexture *QSSGRhiEffectSystem::process(const QSSGRenderLayer &layer,
                                           QRhiTexture *inTexture,
                                           QRhiTexture *inDepthTexture,
                                           QVector2D cameraClipRange)
 {
     QSSG_ASSERT(m_sgContext != nullptr, return inTexture);
+    QSSG_ASSERT(layer.firstEffect != nullptr, return inTexture);
     const auto &rhiContext = m_sgContext->rhiContext();
     const auto &renderer = m_sgContext->renderer();
     QSSG_ASSERT(rhiContext && renderer, return inTexture);
@@ -164,14 +175,17 @@ QRhiTexture *QSSGRhiEffectSystem::process(const QSSGRenderEffect &firstEffect,
     m_depthTexture = inDepthTexture;
     m_cameraClipRange = cameraClipRange;
 
+    const auto viewCount = layer.viewCount;
+
     m_currentUbufIndex = 0;
-    auto *currentEffect = &firstEffect;
+    // FIXME: Keeping the change minimal for now, but we should avoid the need for this cast.
+    QSSGRenderEffect *currentEffect = const_cast<QSSGRenderEffect *>(layer.firstEffect);
     QSSGRhiEffectTexture firstTex{ inTexture, nullptr, nullptr, {}, {}, {} };
-    auto *latestOutput = doRenderEffect(currentEffect, &firstTex);
+    auto *latestOutput = doRenderEffect(currentEffect, &firstTex, viewCount);
     firstTex.texture = nullptr; // make sure we don't delete inTexture when we go out of scope
 
     while ((currentEffect = currentEffect->m_nextEffect)) {
-        auto *effectOut = doRenderEffect(currentEffect, latestOutput);
+        QSSGRhiEffectTexture *effectOut = doRenderEffect(currentEffect, latestOutput, viewCount);
         releaseTexture(latestOutput);
         latestOutput = effectOut;
     }
@@ -203,7 +217,8 @@ QSSGRenderTextureFormat::Format QSSGRhiEffectSystem::overriddenOutputFormat(cons
 }
 
 QSSGRhiEffectTexture *QSSGRhiEffectSystem::doRenderEffect(const QSSGRenderEffect *inEffect,
-                                                          QSSGRhiEffectTexture *inTexture)
+                                                          QSSGRhiEffectTexture *inTexture,
+                                                          quint8 viewCount)
 {
     // Run through the effect commands and render the effect.
     qCDebug(lcEffectSystem) << "START effect " << inEffect->className;
@@ -216,7 +231,7 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::doRenderEffect(const QSSGRenderEffect
 
         switch (theCommand->m_type) {
         case CommandType::AllocateBuffer:
-            allocateBufferCmd(static_cast<QSSGAllocateBuffer *>(theCommand), inTexture, inEffect);
+            allocateBufferCmd(static_cast<QSSGAllocateBuffer *>(theCommand), inTexture, inEffect, viewCount);
             break;
 
         case CommandType::ApplyBufferValue: {
@@ -256,7 +271,7 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::doRenderEffect(const QSSGRenderEffect
         }
 
         case CommandType::BindShader:
-            bindShaderCmd(static_cast<QSSGBindShader *>(theCommand), inEffect);
+            bindShaderCmd(static_cast<QSSGBindShader *>(theCommand), inEffect, viewCount);
             break;
 
         case CommandType::BindTarget: {
@@ -270,13 +285,13 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::doRenderEffect(const QSSGRenderEffect
             qCDebug(lcEffectSystem) << "      Target format override" << QSSGBaseTypeHelpers::toString(f) << "Effective RHI format" << rhiFormat;
             // Make sure we use different names for each effect inside one frame
             QByteArray tmpName = QByteArrayLiteral("__output_").append(QByteArray::number(m_currentUbufIndex));
-            currentOutput = getTexture(tmpName, m_outSize, rhiFormat, true, inEffect);
+            currentOutput = getTexture(tmpName, m_outSize, rhiFormat, true, inEffect, viewCount);
             finalOutputTexture = currentOutput;
             break;
         }
 
         case CommandType::Render:
-            renderCmd(currentInput, currentOutput);
+            renderCmd(currentInput, currentOutput, viewCount);
             currentInput = inTexture; // default input for each new pass is defined to be original input
             break;
 
@@ -290,7 +305,10 @@ QSSGRhiEffectTexture *QSSGRhiEffectSystem::doRenderEffect(const QSSGRenderEffect
     return finalOutputTexture;
 }
 
-void QSSGRhiEffectSystem::allocateBufferCmd(const QSSGAllocateBuffer *inCmd, QSSGRhiEffectTexture *inTexture, const QSSGRenderEffect *inEffect)
+void QSSGRhiEffectSystem::allocateBufferCmd(const QSSGAllocateBuffer *inCmd,
+                                            QSSGRhiEffectTexture *inTexture,
+                                            const QSSGRenderEffect *inEffect,
+                                            quint8 viewCount)
 {
     // Note: Allocate is used both to allocate new, and refer to buffer created earlier
     QSize bufferSize(m_outSize * qreal(inCmd->m_sizeMultiplier));
@@ -299,9 +317,9 @@ void QSSGRhiEffectSystem::allocateBufferCmd(const QSSGAllocateBuffer *inCmd, QSS
     QRhiTexture::Format rhiFormat = (f == QSSGRenderTextureFormat::Unknown) ? inTexture->texture->format()
                                                                             : QSSGBufferManager::toRhiFormat(f);
 
-    QSSGRhiEffectTexture *buf = getTexture(inCmd->m_name, bufferSize, rhiFormat, false, inEffect);
-    auto filter = toRhi(inCmd->m_filterOp);
-    auto tiling = toRhi(inCmd->m_texCoordOp);
+    QSSGRhiEffectTexture *buf = getTexture(inCmd->m_name, bufferSize, rhiFormat, false, inEffect, viewCount);
+    auto filter = QSSGRhiHelpers::toRhi(inCmd->m_filterOp);
+    auto tiling = QSSGRhiHelpers::toRhi(inCmd->m_texCoordOp);
     buf->desc = { filter, filter, QRhiSampler::None, tiling, tiling, QRhiSampler::Repeat };
     buf->flags = inCmd->m_bufferFlags;
 }
@@ -327,12 +345,12 @@ void QSSGRhiEffectSystem::applyInstanceValueCmd(const QSSGApplyInstanceValue *in
                 const QSSGRenderImageTexture texture = theBufferManager->loadRenderImage(image);
                 if (texture.m_texture) {
                     const QSSGRhiSamplerDescription desc{
-                        toRhi(textureProperty.minFilterType),
-                        toRhi(textureProperty.magFilterType),
-                        textureProperty.mipFilterType != QSSGRenderTextureFilterOp::None ? toRhi(textureProperty.mipFilterType) : QRhiSampler::None,
-                        toRhi(textureProperty.horizontalClampType),
-                        toRhi(textureProperty.verticalClampType),
-                        QRhiSampler::Repeat
+                        QSSGRhiHelpers::toRhi(textureProperty.minFilterType),
+                        QSSGRhiHelpers::toRhi(textureProperty.magFilterType),
+                        textureProperty.mipFilterType != QSSGRenderTextureFilterOp::None ? QSSGRhiHelpers::toRhi(textureProperty.mipFilterType) : QRhiSampler::None,
+                        QSSGRhiHelpers::toRhi(textureProperty.horizontalClampType),
+                        QSSGRhiHelpers::toRhi(textureProperty.verticalClampType),
+                        QSSGRhiHelpers::toRhi(textureProperty.zClampType)
                     };
                     addTextureToShaderPipeline(textureProperty.name, texture.m_texture, desc);
                     texAdded = true;
@@ -375,11 +393,12 @@ static const char *effect_builtin_textureMapUVFlipped =
         "    return vec2(uv.x, 1.0 - uv.y);\n"
         "}\n";
 
-QSSGRef<QSSGRhiShaderPipeline> QSSGRhiEffectSystem::buildShaderForEffect(const QSSGBindShader &inCmd,
-                                                                         QSSGProgramGenerator &generator,
-                                                                         QSSGShaderLibraryManager &shaderLib,
-                                                                         QSSGShaderCache &shaderCache,
-                                                                         bool isYUpInFramebuffer)
+QSSGRhiShaderPipelinePtr QSSGRhiEffectSystem::buildShaderForEffect(const QSSGBindShader &inCmd,
+                                                                   QSSGProgramGenerator &generator,
+                                                                   QSSGShaderLibraryManager &shaderLib,
+                                                                   QSSGShaderCache &shaderCache,
+                                                                   bool isYUpInFramebuffer,
+                                                                   int viewCount)
 {
     const auto &key = inCmd.m_shaderPathKey;
     qCDebug(lcEffectSystem) << "    generating new shader pipeline for: " << key;
@@ -404,10 +423,12 @@ QSSGRef<QSSGRhiShaderPipeline> QSSGRhiEffectSystem::buildShaderForEffect(const Q
                                                shaderLib.getShaderMetaData(inCmd.m_shaderPathKey, QSSGShaderCache::ShaderType::Fragment).features,
                                                 shaderLib,
                                                 shaderCache,
-                                                QSSGRhiShaderPipeline::UsedWithoutIa);
+                                                QSSGRhiShaderPipeline::UsedWithoutIa,
+                                                viewCount,
+                                                false);
 }
 
-void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd, const QSSGRenderEffect *inEffect)
+void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd, const QSSGRenderEffect *inEffect, quint8 viewCount)
 {
     QElapsedTimer timer;
     timer.start();
@@ -416,7 +437,8 @@ void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd, const QSSGR
     m_pendingClears.clear();
     m_currentShaderPipeline = nullptr;
 
-    QRhi *rhi = m_sgContext->rhi();
+    const auto &rhiCtx = m_sgContext->rhiContext();
+    QRhi *rhi = rhiCtx->rhi();
     const auto &shaderLib = m_sgContext->shaderLibraryManager();
     const auto &shaderCache = m_sgContext->shaderCache();
 
@@ -467,10 +489,10 @@ void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd, const QSSGR
         // of our local cache (cmd/ubufIndex in cacheKey, not needed here since
         // the result is a new object). Alternatively, the result may be null
         // if there was no hit.
-        QSSGRef<QSSGRhiShaderPipeline> shaderPipeline = shaderCache->tryNewPipelineFromPersistentCache(qsbcKey,
-                                                                                                       inCmd->m_shaderPathKey,
-                                                                                                       features,
-                                                                                                       QSSGRhiShaderPipeline::UsedWithoutIa);
+        const auto &shaderPipeline = shaderCache->tryNewPipelineFromPersistentCache(qsbcKey,
+                                                                                    inCmd->m_shaderPathKey,
+                                                                                    features,
+                                                                                    QSSGRhiShaderPipeline::UsedWithoutIa);
         if (shaderPipeline) {
             m_shaderPipelines.insert(cacheKey, shaderPipeline);
             m_currentShaderPipeline = shaderPipeline.get();
@@ -479,9 +501,10 @@ void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd, const QSSGR
 
     if (!m_currentShaderPipeline) {
         // Final option, generate the shader pipeline
+        Q_TRACE_SCOPE(QSSG_generateShader);
         Q_QUICK3D_PROFILE_START(QQuick3DProfiler::Quick3DGenerateShader);
         const auto &generator = m_sgContext->shaderProgramGenerator();
-        if (auto stages = buildShaderForEffect(*inCmd, *generator, *shaderLib, *shaderCache, rhi->isYUpInFramebuffer())) {
+        if (auto stages = buildShaderForEffect(*inCmd, *generator, *shaderLib, *shaderCache, rhi->isYUpInFramebuffer(), viewCount)) {
             m_shaderPipelines.insert(cacheKey, stages);
             m_currentShaderPipeline = stages.get();
         }
@@ -493,17 +516,17 @@ void QSSGRhiEffectSystem::bindShaderCmd(const QSSGBindShader *inCmd, const QSSGR
     if (m_currentShaderPipeline) {
         const void *cacheKey1 = reinterpret_cast<const void *>(this);
         const void *cacheKey2 = reinterpret_cast<const void *>(qintptr(m_currentUbufIndex));
-        QSSGRhiDrawCallData &dcd = rhiContext->drawCallData({ cacheKey1, cacheKey2, nullptr, 0, QSSGRhiDrawCallDataKey::Effects });
-        m_currentShaderPipeline->ensureCombinedMainLightsUniformBuffer(&dcd.ubuf);
+        QSSGRhiDrawCallData &dcd = QSSGRhiContextPrivate::get(rhiContext.get())->drawCallData({ cacheKey1, cacheKey2, nullptr, 0 });
+        m_currentShaderPipeline->ensureCombinedUniformBuffer(&dcd.ubuf);
         m_currentUBufData = dcd.ubuf->beginFullDynamicBufferUpdateForCurrentFrame();
     } else {
         m_currentUBufData = nullptr;
     }
 
-    rhiContext->stats().registerEffectShaderGenerationTime(timer.elapsed());
+    QSSGRhiContextStats::get(*rhiContext).registerEffectShaderGenerationTime(timer.elapsed());
 }
 
-void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffectTexture *target)
+void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffectTexture *target, quint8 viewCount)
 {
     if (!m_currentShaderPipeline)
         return;
@@ -513,7 +536,9 @@ void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffe
         return;
     }
 
+    // the shader only uses one of these (or none)
     addTextureToShaderPipeline(QByteArrayLiteral("qt_inputTexture"), inTexture->texture, inTexture->desc);
+    addTextureToShaderPipeline(QByteArrayLiteral("qt_inputTextureArray"), inTexture->texture, inTexture->desc);
 
     const auto &rhiContext = m_sgContext->rhiContext();
     const auto &renderer = m_sgContext->renderer();
@@ -529,7 +554,7 @@ void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffe
         // case of outputting into a texture because that will get a clear
         // anyway when rendering the quad.
         if (rt != target->renderTarget) {
-            cb->beginPass(rt, Qt::transparent, { 1.0f, 0 }, nullptr, QSSGRhiContext::commonPassFlags());
+            cb->beginPass(rt, Qt::transparent, { 1.0f, 0 }, nullptr, rhiContext->commonPassFlags());
             QSSGRHICTX_STAT(rhiContext, beginRenderPass(rt));
             cb->endPass();
             QSSGRHICTX_STAT(rhiContext, endRenderPass());
@@ -541,9 +566,11 @@ void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffe
     const QSize outputSize = target->texture->pixelSize();
     addCommonEffectUniforms(inputSize, outputSize);
 
+    QSSGRhiContextPrivate *rhiCtxD = QSSGRhiContextPrivate::get(rhiContext.get());
+
     const void *cacheKey1 = reinterpret_cast<const void *>(this);
     const void *cacheKey2 = reinterpret_cast<const void *>(qintptr(m_currentUbufIndex));
-    QSSGRhiDrawCallData &dcd = rhiContext->drawCallData({ cacheKey1, cacheKey2, nullptr, 0, QSSGRhiDrawCallDataKey::Effects });
+    QSSGRhiDrawCallData &dcd = rhiCtxD->drawCallData({ cacheKey1, cacheKey2, nullptr, 0 });
     dcd.ubuf->endFullDynamicBufferUpdateForCurrentFrame();
     m_currentUBufData = nullptr;
 
@@ -567,11 +594,14 @@ void QSSGRhiEffectSystem::renderCmd(QSSGRhiEffectTexture *inTexture, QSSGRhiEffe
                             rhiContext->sampler(rhiTex.samplerDesc));
     }
     bindings.addUniformBuffer(0, VISIBILITY_ALL, dcd.ubuf);
-    QRhiShaderResourceBindings *srb = rhiContext->srb(bindings);
+
+    QRhiShaderResourceBindings *srb = rhiCtxD->srb(bindings);
 
     QSSGRhiGraphicsPipelineState ps;
     ps.viewport = QRhiViewport(0, 0, float(outputSize.width()), float(outputSize.height()));
-    ps.shaderPipeline = m_currentShaderPipeline;
+    ps.samples = target->renderTarget->sampleCount();
+    ps.viewCount = viewCount;
+    QSSGRhiGraphicsPipelineStatePrivate::setShaderPipeline(ps, m_currentShaderPipeline);
 
     renderer->rhiQuadRenderer()->recordRenderQuadPass(rhiContext.get(), &ps, srb, target->renderTarget, QSSGRhiQuadRenderer::UvCoords);
     m_currentUbufIndex++;
@@ -587,27 +617,27 @@ void QSSGRhiEffectSystem::addCommonEffectUniforms(const QSize &inputSize, const
     QMatrix4x4 mvp;
     if (rhi->isYUpInFramebuffer() != rhi->isYUpInNDC())
         mvp.data()[5] = -1.0f;
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_modelViewProjection", mvp, QSSGRenderShaderDataType::Matrix4x4);
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_modelViewProjection", mvp, QSSGRenderShaderValue::Matrix4x4);
 
     QVector2D size(inputSize.width(), inputSize.height());
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_inputSize", size, QSSGRenderShaderDataType::Vec2);
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_inputSize", size, QSSGRenderShaderValue::Vec2);
 
     size = QVector2D(outputSize.width(), outputSize.height());
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_outputSize", size, QSSGRenderShaderDataType::Vec2);
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_outputSize", size, QSSGRenderShaderValue::Vec2);
 
-    float fc = float(m_sgContext->frameCount());
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_frame_num", fc, QSSGRenderShaderDataType::Float);
+    float fc = float(m_sgContext->renderer()->frameCount());
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_frame_num", fc, QSSGRenderShaderValue::Float);
 
     // Bames and values for uniforms that are also used by default and/or
     // custom materials must always match, effects must not deviate.
 
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_cameraProperties", m_cameraClipRange, QSSGRenderShaderDataType::Vec2);
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_cameraProperties", m_cameraClipRange, QSSGRenderShaderValue::Vec2);
 
     float vp = rhi->isYUpInFramebuffer() ? 1.0f : -1.0f;
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_normalAdjustViewportFactor", vp, QSSGRenderShaderDataType::Float);
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_normalAdjustViewportFactor", vp, QSSGRenderShaderValue::Float);
 
     const float nearClip = rhi->isClipDepthZeroToOne() ? 0.0f : -1.0f;
-    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_nearClipValue", nearClip, QSSGRenderShaderDataType::Float);
+    m_currentShaderPipeline->setUniformValue(m_currentUBufData, "qt_nearClipValue", nearClip, QSSGRenderShaderValue::Float);
 
     if (m_depthTexture) {
         static const QSSGRhiSamplerDescription depthSamplerDesc {
@@ -616,6 +646,7 @@ void QSSGRhiEffectSystem::addCommonEffectUniforms(const QSize &inputSize, const
                     QRhiSampler::ClampToEdge, QRhiSampler::ClampToEdge, QRhiSampler::Repeat
         };
         addTextureToShaderPipeline("qt_depthTexture", m_depthTexture, depthSamplerDesc);
+        addTextureToShaderPipeline("qt_depthTextureArray", m_depthTexture, depthSamplerDesc);
     }
 }
 
